declare namespace SonicDB {
  namespace Utility {
    // validate.ts
    function validatePopulate(obj: PopulateObject): void;
    function validatePopulates(arr: PopulateObject[]): void;
    function validateData(data, schema: Schema): void;
    // util.ts
    function isBrowser(): boolean;
    function isNodejs(): boolean;
    function isInWorker(): boolean;
    function isNestedQuery(spec: any): boolean;
    function isWorkerAvailable(): boolean;
    // schema.ts
    function validateSchema(schemas): void;
    function toRawObject(schemas: SonicDB.Schemas): any;
    function normalizeSchemas(schemas): any;
    // objects.ts
    function wrapObject<T>(obj, collection: CollectionProvider<T> | any): T & DocumentObject;
  }
  export interface DistinctOption {
    format: any;
  }
  export interface DistinctSpec {
    keys: string[];
    option?: DistinctOption;
  }

  type QueryCallable = (db, items: any[], spec: any, queues: any) => Promise<any>;

  namespace Queries {
    function distinct(items: any[], spec: string[] | DistinctSpec): any[];
    function limit(limit: number): any;
    const $filter: QueryCallable;
    const $lookup: QueryCallable
    const $sort: QueryCallable;
    const $limit: QueryCallable;
    const $format: QueryCallable;
    const $distinct: QueryCallable;
    const operators: {
      query: string[];
      queryLogical: string[];
    }
  }

  class ListenerCollection {
    add(key: string, context: any, listener: Function): void;
    insert(indexOrCondition, key, context, listener): void;
    remove(key): void;
    pre(fn): void;
    post(fn): void;
    postFail(fn): void;
    fireAndJoinResults(): Promise<any>;
    fire(...args): Promise<any>;
  }

  interface PopulateObject {
    localField: string;
    foreignField: string;
    ref: string;
    as?: string;
  }
  interface DocumentObjectCondition {
    id?: string;
    [key: string]: any;
  }

  interface DocumentObject {
    id?: any;
    updatedAt?: number | Date;
    createdAt?: number | Date;
    update?: (data) => Promise<void>;
    delete?: () => Promise<void>;
  }


  interface EventOptions {
    /**
     * The event name for the socket to listen when server send an event to notify 
     * that there is a data changed on the server, this was also being used in collection
     * to notify user that there is some change made in the collection.
     */
    changeEventName?: string;
    /**
     * The event name for the socket to listen when server send an event to notify 
     * that there is a data removed on the server, this was also being used in collection
     * to notify user that there is some change made in the collection. 
     */
    removeEventName?: string;
    /**
     * The event name for the socket to listen when server send an event to notify 
     * that there is a new data added on the server, this was also being used in collection
     * to notify user that there is some change made in the collection.
     */
    addEventName?: string;
    /**
     * The event name for the socket to listen when server send all data to be fetched 
     * in the client
     */
    fetchEventName?: string;
  }
  abstract class CollectionProvider<T> extends EventEmitter {
    abstract find(obj: any): CursorProvider<T>;
    abstract findOne(obj: any): CursorProvider<T>;
    abstract add(data: T): Promise<any>;
    abstract update(condition: T & SonicDB.DocumentObjectCondition, data: any): Promise<any>;
    abstract remove(condition: T & SonicDB.DocumentObjectCondition): Promise<any>;
    abstract drop(): Promise<any>;
  }
  abstract class CursorProvider<T> {
    abstract get(): Promise<T & DocumentObject>;
    abstract toArray(): Promise<(T & DocumentObject)[]>;
    abstract filter(spec: any): CursorProvider<T>;
    abstract limit(num: number): CursorProvider<T>;
    abstract sort(spec: any): CursorProvider<T>;
    abstract populate(spec: SonicDB.PopulateObject): CursorProvider<T>;
    abstract populates(spec: SonicDB.PopulateObject[]): CursorProvider<T>;
  }

  abstract class DatabaseProvider extends EventEmitter {
    name: string;
    options: any;
    syncAdapter: SonicDB.SyncProvider;
    private _context: SonicDB;
    public abstract context: SonicDB;
    abstract collection<T>(name: string): CollectionProvider<T>;
    abstract close(): void;
    abstract drop(name?: string): Promise<any>;
    abstract readonly isOpen: boolean;
  }

  abstract class SyncProvider extends EventEmitter {
    dbName: string;
    schemas: any;
    localAdapter: DatabaseProvider;
    context: SonicDB;
    abstract nextId(name: string): any;
    abstract start(): void;
    abstract stop(): void;
    abstract authenticate(opts: SonicDB.AuthOptions): Promise<any>;
    abstract add(name: string, data: any): Promise<any>;
    abstract update(name: string, condition, data: any): Promise<any>;
    abstract remove(name: string, condition): Promise<any>;
    abstract drop(name?: string): Promise<any>;
  }

  interface Queue {
    id?: any;
    tableName?: string;
    type?: string;
    refId?: string,
    data?: any;
  }

  interface SyncOptions {
    url?: string;
    enabled?: boolean;
    connectionType?: string;
    events?: EventOptions;
    adapter?: SyncProvider;
    autoStart?: boolean;
    interval?: number;
  }

  interface SonicDBOptions {
    adapter: DatabaseProvider;
    sync?: SyncOptions;
  }

  interface SchemaColumnProps {
    autoIncrement?: boolean;
    primaryKey?: boolean;
    unique?: boolean;
    type: typeof String
    | typeof Number
    | typeof Array
    | typeof Date
    | typeof Object;
  }
  interface SchemaProps {
    [name: string]: SchemaColumnProps
    | typeof String
    | typeof Number
    | typeof Array
    | typeof Date
    | typeof Object;
  }
  interface Schemas {
    [name: string]: Schema;
  }
  interface SchemaOptions {
    timestamp?: boolean;
  }
  interface AuthOptions {
    username?: string;
    email?: string;
    password: string;
  }
  class Schema {
    constructor(obj?: SchemaProps, options?: SchemaOptions);
    options: SchemaOptions;
    raw: string;
    primaryKeyField: string;
    fields: SchemaProps;
    withTimestamp: boolean;
    autoIncrementEnabled: boolean;
  }

  class Table<T> {
    select(q?: string[] | string);
  }
  class NetworkProvider extends NodeJS.EventEmitter {
    connected: boolean;
  }
  class EventEmitter extends NodeJS.EventEmitter {

  }
}

declare class SonicDB extends NodeJS.EventEmitter {
  syncAdapter: SonicDB.SyncProvider;
  constructor(options: SonicDB.SonicDBOptions);
  collection<T>(name: string): SonicDB.CollectionProvider<T>;
  table<T>(name: string): SonicDB.Table<T>;
  close(): void;
  drop(name?: string): any;
  authenticate(opts: SonicDB.AuthOptions): Promise<any>;
  readonly isOpen: boolean;
}

declare module 'sonicdb' {
  export = SonicDB;
}
